@@ -1,7 +1,7 @@
 #ifndef _MIBGROUP_DISKIO_H
 #define _MIBGROUP_DISKIO_H
 
-config_require(util_funcs)
+config_require(util_funcs/header_simple_table)
 config_add_mib(UCD-DISKIO-MIB)
 
     /*
@@ -32,5 +32,7 @@ config_add_mib(UCD-DISKIO-MIB)
 #define DISKIO_LA1		9
 #define DISKIO_LA5              10
 #define DISKIO_LA15             11
+#define DISKIO_NREADX           12
+#define DISKIO_NWRITTENX        13
 
 #endif                          /* _MIBGROUP_DISKIO_H */
